# Multisign wallet contract for TON Blockchain

## Smart Contract Features

- Autocleanup of expired orders;
- Signs whole message data (sequence number, expired date, send mode and internal message);
- Can add signatures without original message. (i.e. reading pending orders from blockchain and send just missing signatures);
- Public methods to inspect pending orders and to get a hash of order message data;
- Signatures can be collected offline (see utility script files);

The code is split in different files. Each script file has usage description built in.
The smart contract itself is written in `FunC` and should be compiled before uploading into the blockchain.

## Main files

| Filename                 | Description                                                                               |
| ------------------------ | ----------------------------------------------------------------------------------------- |
| multisign-wallet-code.fc | Contact code in FunC                                                                      |
| new-multising-wallet.fif | Script that creates query for new multisign wallet                                        |
| multisign-wallet.fif     | Script that creates query request to multisign wallet created by new-multisign-wallet.fif |

## Utility script files

| Filename           | Description                                            |
| ------------------ | ------------------------------------------------------ |
| add-signatures.fif | Add signatures to existing query file                  |
| inspect-query.fif  | Show multisign wallet query parameters from query file |
| merge-queries.fif  | Merge signatures from two query files                  |

## Internal files

| Filename                    | Description                                              |
| --------------------------- | -------------------------------------------------------- |
| lib/MultiSignWalletUtil.fif | Various functions to work with messages                  |
| multisign-wallet-test.fif   | Contract integration tests using `runvmctx` fift command |

## HOWTO

You can set the FIFTPATH environment variable to `<source-directory>/crypto/fift/lib:<source-directory>/crypto/smartcont` (you can read about this in [TON HOWTO](https://test.ton.org/HOWTO.txt)).

To send queries you can use lite-client by `sendfile <file-query>.boc` command.

### 1. Create multisign wallet

#### Build contract code:

```sh
crypto/func -PAo multisign-wallet-code.fif <source-directory>/crypto/smartcont/stdlib.fc multisign-wallet-code.fc
```

This command will create `multisign-wallet-code.fc` which will be used in next step.

#### Create message to make new multisign wallet:

```sh
crypto/fift -I<source-directory>/crypto/fift/lib -s new-multisign-wallet.fif <required number of keys> <total keys> <workchain-id> [<filename-base>]
```

This command will create `<total keys>` keypairs (or load them if already exist), `<filename-base>.addr` with address of the wallet
and `<filename-base-query>.boc` query file which can be sent into network.

### 2. Create transfer Grams query

```sh
crypto/fift -I<source-directory>/crypto/fift/lib -s multisign-wallet.fif <wallet-addr-file> <dest-addr> <seqno> <amount> <expire> <filename> [<filename> ...]
```

This command will create `multisign-wallet-query.boc` query file which can be sent into network.
If amount of signatures is more or equal to _k_ then the transfer will happen immediately otherwise
it will be stored until more signatures arrive or expire time passed.

### 3. Add signatures to pending order from blockchain

If one has private key from created multisign wallet it is possible to add its signatures without original message.
To do it get message data signature using public method:

```sh
lite-client/lite-client -C ton-lite-client-test1.config.json -rc 'runmethod <multisignatur-wallet-address> pending_order_by_seq_no <order-seqno>'
```

At the end of the output the message data hash will be in the for of:

```
...
result:  [ <message-data-hash> ]
```

Use this value and signature(s) from `<filename> [<filename> ...]` in `sign-data.fif`:

```sh
crypto/fift -I<source-directory>/crypto/fift/lib -s sign-data.fif <multisignature-wallet-address> <order-seqno> <message-data-hash> <filename> [<filename> ...]
```

The result will be `sign-data-out.boc` query file which can be sent into network.

## Contract public methods

> To call public methods you can use lite-client `runmethod <contract-address> <method> <args>` command.

- **pending_orders** - get all pending orders sequence numbers in list;
- **pending_orders_by_public_key** - get status (signed or not) of all pending orders by public key;
- **pending_order_by_seq_no** - get order message hash by sequence number _(used to send missing signatures to order_);
